/**
 * Copyright (c) 2017-2018 Thales Global Services S.A.S.
 * This program and the accompanying materials are made
 * available under the terms of the Eclipse Public License 2.0
 * which is available at https://www.eclipse.org/legal/epl-2.0/
 * 
 * SPDX-License-Identifier: EPL-2.0
 * 
 * Contributors:
 *     Abel Hegedus (IncQuery Labs Ltd.) - initial API and implementation
 */
package org.eclipse.emf.diffmerge.patch.persistence.emf.transform;

import com.google.common.collect.Iterables;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.function.Consumer;
import org.apache.log4j.Logger;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.diffmerge.patch.AbstractModelPatchEntry;
import org.eclipse.emf.diffmerge.patch.AttributeEntry;
import org.eclipse.emf.diffmerge.patch.ElementEntry;
import org.eclipse.emf.diffmerge.patch.ReferenceEntry;
import org.eclipse.emf.diffmerge.patch.StructuralFeatureEntry;
import org.eclipse.emf.diffmerge.patch.api.ChangeDirection;
import org.eclipse.emf.diffmerge.patch.api.Identifiable;
import org.eclipse.emf.diffmerge.patch.api.ModelPatch;
import org.eclipse.emf.diffmerge.patch.api.ModelPatchMetadata;
import org.eclipse.emf.diffmerge.patch.persistence.emf.emodelpatch.EAttributeEntry;
import org.eclipse.emf.diffmerge.patch.persistence.emf.emodelpatch.EChangeDirection;
import org.eclipse.emf.diffmerge.patch.persistence.emf.emodelpatch.EElementEntry;
import org.eclipse.emf.diffmerge.patch.persistence.emf.emodelpatch.EIdentifiable;
import org.eclipse.emf.diffmerge.patch.persistence.emf.emodelpatch.EModelPatch;
import org.eclipse.emf.diffmerge.patch.persistence.emf.emodelpatch.EModelPatchEMFMetadata;
import org.eclipse.emf.diffmerge.patch.persistence.emf.emodelpatch.EModelPatchEntry;
import org.eclipse.emf.diffmerge.patch.persistence.emf.emodelpatch.EModelPatchFactory;
import org.eclipse.emf.diffmerge.patch.persistence.emf.emodelpatch.EModelPatchMetadata;
import org.eclipse.emf.diffmerge.patch.persistence.emf.emodelpatch.EModelPatchPackage;
import org.eclipse.emf.diffmerge.patch.persistence.emf.emodelpatch.EReferenceEntry;
import org.eclipse.emf.diffmerge.patch.persistence.emf.emodelpatch.EStructuralFeatureEntry;
import org.eclipse.emf.diffmerge.patch.runtime.EMFModelPatchMetadata;
import org.eclipse.emf.ecore.EAttribute;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.eclipse.viatra.query.runtime.base.api.BaseIndexOptions;
import org.eclipse.viatra.query.runtime.base.api.IndexingLevel;
import org.eclipse.viatra.query.runtime.base.api.NavigationHelper;
import org.eclipse.viatra.query.runtime.base.api.ViatraBaseFactory;
import org.eclipse.xtext.xbase.lib.CollectionLiterals;
import org.eclipse.xtext.xbase.lib.Exceptions;
import org.eclipse.xtext.xbase.lib.Extension;
import org.eclipse.xtext.xbase.lib.Functions.Function1;
import org.eclipse.xtext.xbase.lib.IterableExtensions;
import org.eclipse.xtext.xbase.lib.ObjectExtensions;
import org.eclipse.xtext.xbase.lib.Procedures.Procedure1;

@SuppressWarnings("all")
public class ModelPatchToEMF {
  @Extension
  private final EModelPatchFactory factory = EModelPatchFactory.eINSTANCE;
  
  private final EAttribute targetIdentiferFeature = EModelPatchPackage.eINSTANCE.getEIdentifiable_TargetIdentifier();
  
  private final EReference elementsFeature = EModelPatchPackage.eINSTANCE.getEModelPatch_Elements();
  
  private final EReference typesFeature = EModelPatchPackage.eINSTANCE.getEModelPatch_Types();
  
  private final EReference featuresFeature = EModelPatchPackage.eINSTANCE.getEModelPatch_Features();
  
  private final NavigationHelper indexer;
  
  private final ModelPatch source;
  
  private final EModelPatch target;
  
  private final Logger logger;
  
  /**
   * When true, the same targetIdentifier can appear in the elements, types and features lists.
   * 
   * Note that duplicates in the same list are not allowed even if set to true.
   */
  private final boolean allowDuplicateIdentifiables;
  
  public ModelPatchToEMF(final ModelPatch source, final EModelPatch target, final Logger logger, final boolean allowDuplicateIdentifiables) {
    try {
      this.source = source;
      this.target = target;
      this.allowDuplicateIdentifiables = allowDuplicateIdentifiables;
      this.logger = logger;
      ViatraBaseFactory _instance = ViatraBaseFactory.getInstance();
      BaseIndexOptions _baseIndexOptions = new BaseIndexOptions(false, IndexingLevel.NONE);
      this.indexer = _instance.createNavigationHelper(target, _baseIndexOptions, logger);
      this.indexer.registerEStructuralFeatures(Collections.<EStructuralFeature>unmodifiableSet(CollectionLiterals.<EStructuralFeature>newHashSet(this.targetIdentiferFeature)), IndexingLevel.FULL);
    } catch (Throwable _e) {
      throw Exceptions.sneakyThrow(_e);
    }
  }
  
  public void transform() {
    this.transformElement(this.source.getMetadata());
    final Consumer<AbstractModelPatchEntry> _function = (AbstractModelPatchEntry it) -> {
      this.transformElement(it);
    };
    Iterables.<AbstractModelPatchEntry>filter(this.source.getEntries(), AbstractModelPatchEntry.class).forEach(_function);
    this.indexer.dispose();
  }
  
  protected void transformElement(final ModelPatchMetadata metadata) {
    EModelPatchMetadata _xifexpression = null;
    if ((metadata instanceof EMFModelPatchMetadata)) {
      EModelPatchEMFMetadata _createEModelPatchEMFMetadata = this.factory.createEModelPatchEMFMetadata();
      final Procedure1<EModelPatchEMFMetadata> _function = (EModelPatchEMFMetadata it) -> {
        EList<String> _modelUris = it.getModelUris();
        List<String> _modelUriList = ((EMFModelPatchMetadata)metadata).getModelUriList();
        Iterables.<String>addAll(_modelUris, _modelUriList);
        EList<String> _usedNamespaceUris = it.getUsedNamespaceUris();
        List<String> _usedNamespaceUris_1 = ((EMFModelPatchMetadata)metadata).getUsedNamespaceUris();
        Iterables.<String>addAll(_usedNamespaceUris, _usedNamespaceUris_1);
      };
      _xifexpression = ObjectExtensions.<EModelPatchEMFMetadata>operator_doubleArrow(_createEModelPatchEMFMetadata, _function);
    } else {
      _xifexpression = this.factory.createEModelPatchMetadata();
    }
    final EModelPatchMetadata emetadata = _xifexpression;
    final Procedure1<EModelPatchMetadata> _function_1 = (EModelPatchMetadata it) -> {
      it.setAuthor(metadata.getAuthor());
      it.setCreatedAt(metadata.getCreatedAt());
      it.setDescription(metadata.getDescription());
      it.setInput(metadata.getInput());
    };
    ObjectExtensions.<EModelPatchMetadata>operator_doubleArrow(emetadata, _function_1);
    this.target.setMetadata(emetadata);
    return;
  }
  
  protected void transformElement(final AbstractModelPatchEntry entry) {
    final EModelPatchEntry eentry = this.transformEntry(entry);
    this.target.getEntries().add(eentry);
    return;
  }
  
  protected EModelPatchEntry _transformEntry(final ElementEntry entry) {
    EElementEntry _createEElementEntry = this.factory.createEElementEntry();
    final Procedure1<EElementEntry> _function = (EElementEntry it) -> {
      this.transformAbstractEntryFeatures(entry, it);
      it.setType(this.ensureEIdentifiable(entry.getType(), this.typesFeature));
    };
    return ObjectExtensions.<EElementEntry>operator_doubleArrow(_createEElementEntry, _function);
  }
  
  protected EModelPatchEntry _transformEntry(final AttributeEntry entry) {
    EAttributeEntry _createEAttributeEntry = this.factory.createEAttributeEntry();
    final Procedure1<EAttributeEntry> _function = (EAttributeEntry it) -> {
      this.transformStructuralFeatureEntryFeatures(entry, it);
      it.setValue(entry.getValue());
    };
    return ObjectExtensions.<EAttributeEntry>operator_doubleArrow(_createEAttributeEntry, _function);
  }
  
  protected EModelPatchEntry _transformEntry(final ReferenceEntry entry) {
    EReferenceEntry _createEReferenceEntry = this.factory.createEReferenceEntry();
    final Procedure1<EReferenceEntry> _function = (EReferenceEntry it) -> {
      this.transformStructuralFeatureEntryFeatures(entry, it);
      it.setTarget(this.ensureEIdentifiable(entry.getTarget(), this.elementsFeature));
    };
    return ObjectExtensions.<EReferenceEntry>operator_doubleArrow(_createEReferenceEntry, _function);
  }
  
  protected void transformAbstractEntryFeatures(final AbstractModelPatchEntry entry, final EModelPatchEntry eentry) {
    eentry.setDirection(this.transformDirection(entry.getDirection()));
    eentry.setContext(this.ensureEIdentifiable(entry.getContext(), this.elementsFeature));
  }
  
  protected void transformStructuralFeatureEntryFeatures(final StructuralFeatureEntry entry, final EStructuralFeatureEntry eentry) {
    this.transformAbstractEntryFeatures(entry, eentry);
    eentry.setFeature(this.ensureEIdentifiable(entry.getFeature(), this.featuresFeature));
  }
  
  protected EChangeDirection transformDirection(final ChangeDirection direction) {
    EChangeDirection _switchResult = null;
    if (direction != null) {
      switch (direction) {
        case ADD:
          _switchResult = EChangeDirection.ADD;
          break;
        case REMOVE:
          _switchResult = EChangeDirection.REMOVE;
          break;
        default:
          break;
      }
    }
    return _switchResult;
  }
  
  protected EIdentifiable ensureEIdentifiable(final Identifiable identifiable, final EStructuralFeature identifiableContainmentFeature) {
    final String identifier = identifiable.getIdentifier();
    final Iterable<EIdentifiable> allEIdentifiables = Iterables.<EIdentifiable>filter(this.indexer.findByAttributeValue(identifier, this.targetIdentiferFeature), EIdentifiable.class);
    boolean _isEmpty = IterableExtensions.isEmpty(allEIdentifiables);
    if (_isEmpty) {
      final EIdentifiable eidentifable = this.prepareEIdentifiable(identifiable, identifiableContainmentFeature);
      return eidentifable;
    }
    final int noOfIdentifiables = IterableExtensions.size(allEIdentifiables);
    final Function1<EIdentifiable, Boolean> _function = (EIdentifiable it) -> {
      EReference _eContainmentFeature = it.eContainmentFeature();
      return Boolean.valueOf((_eContainmentFeature == identifiableContainmentFeature));
    };
    final Iterable<EIdentifiable> onlyGivenFeature = IterableExtensions.<EIdentifiable>filter(allEIdentifiables, _function);
    final int noOfIdentifiablesInGivenFeature = IterableExtensions.size(onlyGivenFeature);
    if (((noOfIdentifiablesInGivenFeature == 1) && (noOfIdentifiables == 1))) {
      return IterableExtensions.<EIdentifiable>head(onlyGivenFeature);
    }
    if ((this.allowDuplicateIdentifiables && IterableExtensions.isEmpty(onlyGivenFeature))) {
      final EIdentifiable eidentifable_1 = this.prepareEIdentifiable(identifiable, identifiableContainmentFeature);
      return eidentifable_1;
    }
    String _format = String.format("Duplicate identifiers found: %s", identifier);
    throw new IllegalStateException(_format);
  }
  
  protected EIdentifiable prepareEIdentifiable(final Identifiable identifiable, final EStructuralFeature identifiableContainmentFeature) {
    EIdentifiable _xblockexpression = null;
    {
      EIdentifiable _createEIdentifiable = this.factory.createEIdentifiable();
      final Procedure1<EIdentifiable> _function = (EIdentifiable it) -> {
        it.setTargetIdentifier(identifiable.getIdentifier());
        it.setDescription(identifiable.getDescription());
      };
      final EIdentifiable eidentifable = ObjectExtensions.<EIdentifiable>operator_doubleArrow(_createEIdentifiable, _function);
      Object _eGet = this.target.eGet(identifiableContainmentFeature);
      final List<EIdentifiable> list = ((List<EIdentifiable>) _eGet);
      list.add(eidentifable);
      _xblockexpression = eidentifable;
    }
    return _xblockexpression;
  }
  
  protected EModelPatchEntry transformEntry(final AbstractModelPatchEntry entry) {
    if (entry instanceof AttributeEntry) {
      return _transformEntry((AttributeEntry)entry);
    } else if (entry instanceof ReferenceEntry) {
      return _transformEntry((ReferenceEntry)entry);
    } else if (entry instanceof ElementEntry) {
      return _transformEntry((ElementEntry)entry);
    } else {
      throw new IllegalArgumentException("Unhandled parameter types: " +
        Arrays.<Object>asList(entry).toString());
    }
  }
}
