/**
 * Copyright (c) 2016-2018 Thales Global Services S.A.S.
 * This program and the accompanying materials are made
 * available under the terms of the Eclipse Public License 2.0
 * which is available at https://www.eclipse.org/legal/epl-2.0/
 * 
 * SPDX-License-Identifier: EPL-2.0
 * 
 * Contributors:
 *   Abel Hegedus, Tamas Borbas, Peter Lunk (IncQuery Labs Ltd.) - initial API and implementation
 */
package org.eclipse.emf.diffmerge.patch.ui.handlers;

import com.google.common.base.Objects;
import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.emf.common.util.EMap;
import org.eclipse.emf.diffmerge.api.IMatch;
import org.eclipse.emf.diffmerge.api.Role;
import org.eclipse.emf.diffmerge.api.diff.IDifference;
import org.eclipse.emf.diffmerge.api.diff.IPresenceDifference;
import org.eclipse.emf.diffmerge.diffdata.EMatch;
import org.eclipse.emf.diffmerge.ui.diffuidata.ComparisonSelection;
import org.eclipse.emf.diffmerge.ui.viewers.EMFDiffNode;
import org.eclipse.emf.diffmerge.ui.viewers.MergeImpactViewer;
import org.eclipse.jface.operation.IRunnableWithProgress;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.progress.IProgressService;
import org.eclipse.xtext.xbase.lib.Exceptions;

@SuppressWarnings("all")
public class ComparisonSelectionUtil {
  protected Collection<IDifference> getDiffsToMerge(final ComparisonSelection selection, final boolean toLeft_p, final EMFDiffNode input) {
    final Role destination = Role.REFERENCE;
    final boolean isCoverChildren = true;
    final boolean isIncrementalMode = false;
    final List<EMatch> selectedMatches = this.getSelectedMatchesForInteractions(selection);
    List<IDifference> _xifexpression = null;
    boolean _isEmpty = selectedMatches.isEmpty();
    boolean _not = (!_isEmpty);
    if (_not) {
      _xifexpression = this.getDifferencesToMerge(selectedMatches, destination, isCoverChildren, isIncrementalMode, input);
    } else {
      _xifexpression = input.getCategoryManager().getPendingDifferencesFiltered(selection.asDifferencesToMerge());
    }
    final Collection<IDifference> toMerge = _xifexpression;
    final MergeImpactViewer.ImpactInput mergeInput = new MergeImpactViewer.ImpactInput(toMerge, toLeft_p, input);
    final IProgressService progress = PlatformUI.getWorkbench().getProgressService();
    try {
      progress.busyCursorWhile(new IRunnableWithProgress() {
        /**
         * @see IRunnableWithProgress#run(IProgressMonitor)
         */
        @Override
        public void run(final IProgressMonitor monitor_p) throws InvocationTargetException, InterruptedException {
          mergeInput.compute(monitor_p);
        }
      });
    } catch (final Throwable _t) {
      if (_t instanceof Exception) {
        return null;
      } else {
        throw Exceptions.sneakyThrow(_t);
      }
    }
    final EMap<IMatch, Collection<IDifference>> explicitImpacts = mergeInput.getImpact(true);
    final EMap<IMatch, Collection<IDifference>> implicitImpacts = mergeInput.getImpact(false);
    final Set<IDifference> diffs = new HashSet<IDifference>();
    for (final Map.Entry<IMatch, Collection<IDifference>> impact : implicitImpacts) {
      diffs.addAll(impact.getValue());
    }
    for (final Map.Entry<IMatch, Collection<IDifference>> impact_1 : explicitImpacts) {
      diffs.addAll(impact_1.getValue());
    }
    return diffs;
  }
  
  protected List<EMatch> getSelectedMatchesForInteractions(final ComparisonSelection selection_p) {
    List<EMatch> selectedMatches = selection_p.getSelectedMatches();
    boolean _isEmpty = selectedMatches.isEmpty();
    if (_isEmpty) {
      final List<EMatch> treePath = selection_p.getSelectedTreePath();
      boolean _isEmpty_1 = treePath.isEmpty();
      boolean _not = (!_isEmpty_1);
      if (_not) {
        int _size = treePath.size();
        int _minus = (_size - 1);
        selectedMatches = Collections.<EMatch>singletonList(treePath.get(_minus));
      }
    }
    return selectedMatches;
  }
  
  /**
   * Copied from {@link org.eclipse.emf.diffmerge.ui.viewers.ComparisonViewer}
   * 
   * Return the differences to merge from a given list of selected matches and the given
   * criteria
   * @param selectedMatches_p a non-null list
   * @param coverChildren_p whether children of the matches must be covered
   * @param incrementalMode_p whether optional deletions must be skipped
   * @return a non-null, potentially empty, unmodifiable list
   */
  protected List<IDifference> getDifferencesToMerge(final List<EMatch> selectedMatches_p, final Role destination_p, final boolean coverChildren_p, final boolean incrementalMode_p, final EMFDiffNode input) {
    final List<IDifference> result = new ArrayList<IDifference>();
    final IProgressService progress = PlatformUI.getWorkbench().getProgressService();
    try {
      progress.busyCursorWhile(new IRunnableWithProgress() {
        /**
         * @see IRunnableWithProgress#run(IProgressMonitor)
         */
        @Override
        public void run(final IProgressMonitor monitor_p) throws InvocationTargetException, InterruptedException {
          for (final EMatch selectedMatch : selectedMatches_p) {
            if (coverChildren_p) {
              ComparisonSelectionUtil.this.addDifferencesToMergeRec(result, selectedMatch, destination_p, incrementalMode_p, input);
            } else {
              ComparisonSelectionUtil.this.addDifferencesToMerge(result, selectedMatch, destination_p, incrementalMode_p, input);
            }
          }
        }
      });
    } catch (final Throwable _t) {
      if (_t instanceof Exception) {
      } else {
        throw Exceptions.sneakyThrow(_t);
      }
    }
    return Collections.<IDifference>unmodifiableList(result);
  }
  
  /**
   * Copied from {@link org.eclipse.emf.diffmerge.ui.viewers.ComparisonViewer}
   * 
   * Add differences to merge on the given match to the given list according
   * to the given criteria
   * @param toMerge_p a non-null, modifiable list
   * @param match_p a non-null match
   * @param destination_p a non-null role which is TARGET or REFERENCE
   * @param incrementalMode_p whether optional deletions must be skipped
   */
  protected void addDifferencesToMerge(final List<IDifference> toMerge_p, final IMatch match_p, final Role destination_p, final boolean incrementalMode_p, final EMFDiffNode input) {
    List<IDifference> _allDifferences = match_p.getAllDifferences();
    for (final IDifference difference : _allDifferences) {
      boolean _isFiltered = input.getCategoryManager().isFiltered(difference);
      boolean _not = (!_isFiltered);
      if (_not) {
        if (((!incrementalMode_p) || ((difference instanceof IPresenceDifference) && 
          (!Objects.equal(((IPresenceDifference) difference).getPresenceRole(), destination_p))))) {
          toMerge_p.add(difference);
        }
      }
    }
  }
  
  /**
   * Copied from {@link org.eclipse.emf.diffmerge.ui.viewers.ComparisonViewer}
   * 
   * Add differences to merge on the given match and its children to the given list according
   * to the given criteria
   * @param toMerge_p a non-null, modifiable list
   * @param match_p a non-null match
   * @param destination_p a non-null role which is TARGET or REFERENCE
   * @param incrementalMode_p whether optional deletions must be skipped
   */
  protected void addDifferencesToMergeRec(final List<IDifference> toMerge_p, final IMatch match_p, final Role destination_p, final boolean incrementalMode_p, final EMFDiffNode input) {
    this.addDifferencesToMerge(toMerge_p, match_p, destination_p, incrementalMode_p, input);
    List<IMatch> _childrenForMerge = input.getCategoryManager().getChildrenForMerge(match_p);
    for (final IMatch child : _childrenForMerge) {
      this.addDifferencesToMergeRec(toMerge_p, child, destination_p, incrementalMode_p, input);
    }
  }
}
